@@ -42,7 +42,7 @@ namespace Mono.WebServer
 	//
 	internal class ModMonoWorker: Worker
 	{
-		public LingeringNetworkStream Stream;
+		public NetworkStream Stream;
 		
 		ApplicationServer server;
 		ModMonoRequest modRequest;
@@ -53,8 +53,7 @@ namespace Mono.WebServer
 		
 		public ModMonoWorker (Socket client, ApplicationServer server)
 		{
-			Stream = new LingeringNetworkStream (client, true);
-			Stream.EnableLingering = false;
+			Stream = new NetworkStream (client, true);
 			this.client = client;
 			this.server = server;
 		}
@@ -69,7 +68,8 @@ namespace Mono.WebServer
 
 				// IOException, like EndOfStreamException, might be ok.
 
-				Console.Error.WriteLine (e);
+				if (!(e is EndOfStreamException))
+					Console.Error.WriteLine (e);
 
 				try {
 					// Closing is enough for mod_mono. the module will return a 50x
@@ -78,7 +78,7 @@ namespace Mono.WebServer
 						Stream = null;
 					}
 				} catch {}
-				if (broker != null && requestId != -1) {
+				if (!server.SingleApplication && broker != null && requestId != -1) {
 					broker.UnregisterRequest (requestId);
 					requestId = -1;
 				}
@@ -202,10 +202,11 @@ namespace Mono.WebServer
 			}
 			modRequest = rr.Request;
 			
-			broker = (ModMonoRequestBroker) vapp.RequestBroker;
-			broker.UnregisterRequestEvent += new BaseRequestBroker.UnregisterRequestEventHandler (OnUnregisterRequest);
-			
-			requestId = broker.RegisterRequest (this);
+			if (!server.SingleApplication) {
+				broker = (ModMonoRequestBroker) vapp.RequestBroker;
+				broker.UnregisterRequestEvent += new BaseRequestBroker.UnregisterRequestEventHandler (OnUnregisterRequest);
+				requestId = broker.RegisterRequest (this);
+			}
 			
 			host.ProcessRequest (requestId, 
 					     modRequest.GetHttpVerbName(), 
@@ -218,7 +219,8 @@ namespace Mono.WebServer
 					     modRequest.GetRemotePort(), 
 					     modRequest.GetRemoteName(), 
 					     modRequest.GetAllHeaders(),
-					     modRequest.GetAllHeaderValues());
+					     modRequest.GetAllHeaderValues(),
+					     (requestId == -1) ? this : null);
 		}
 
 		void OnUnregisterRequest (object sender, UnregisterRequestEventArgs args)
@@ -294,5 +296,9 @@ namespace Mono.WebServer
 		{
 			modRequest.SetOutputBuffering (doBuffer);
 		}
+
+		public bool HeadersSent {
+			get { return modRequest.HeadersSent; }
+		}
 	}	
 }
