@@ -56,10 +56,11 @@ namespace Mono.WebServer
 		PrivateKeySelectionCallback keyCB;
 		bool allowClientCert;
 		bool requireClientCert;
+		bool single_app;
 
 		public XSPWebSource(IPAddress address, int port, SecurityProtocolType securityProtocol,
 				    X509Certificate cert, PrivateKeySelectionCallback keyCB, 
-				    bool allowClientCert, bool requireClientCert)
+				    bool allowClientCert, bool requireClientCert, bool single_app)
 		{			
 			secureConnection = (cert != null && keyCB != null);
 			this.bindAddress = new IPEndPoint (address, port);
@@ -68,13 +69,19 @@ namespace Mono.WebServer
 			this.keyCB = keyCB;
 			this.allowClientCert = allowClientCert;
 			this.requireClientCert = requireClientCert;
+			this.single_app = single_app;
 		}
 
-		public XSPWebSource (IPAddress address, int port)
+		public XSPWebSource (IPAddress address, int port, bool single_app)
 		{
+			this.single_app = single_app;
 			SetListenAddress (address, port);
 		}
-		
+
+		public XSPWebSource (IPAddress address, int port) : this (address, port, false)
+		{
+		}
+
 		public void SetListenAddress (int port)
 		{
 			SetListenAddress (IPAddress.Any, port);
