@@ -99,7 +99,7 @@ namespace Mono.WebServer
 
 			int platform = (int) Environment.OSVersion.Platform;
 			string plat;
-			if (platform == 4 || platform == 128)
+			if (platform == 4 || platform == 6 || platform == 128)
 				plat = "Unix";
 			else
 				plat = ((PlatformID) platform).ToString ();
@@ -177,7 +177,7 @@ namespace Mono.WebServer
 			this.remoteEP = remoteEP;
 			this.verb = verb;
 			try {
-				Paths.GetPathsFromUri (path, out this.path, out pathInfo);
+				Paths.GetPathsFromUri (appHost, verb, path, out this.path, out pathInfo);
 			} catch {
 				CloseConnection ();
 				throw;
@@ -775,7 +775,7 @@ namespace Mono.WebServer
 					int n = send (socket, ptr + total, (IntPtr) (len - total), (IntPtr) 0x4000);
 					if (n >= 0) {
 						total += n;
-					} else {
+					} else if (Marshal.GetLastWin32Error () != 4 /* EINTR */) {
 						throw new IOException ();
 					}
 				}
